@@ -5,25 +5,26 @@
 Generic Job class extends the JobCore class with all the functionality to run the job object.
 """
 
-# import copy
 import signal
 from datetime import datetime
 import os
 
-# import sys
 import posixpath
-import psutil
 import multiprocessing
 from pyiron_base.job.wrapper import JobWrapper
 from pyiron_base.settings.generic import Settings
 from pyiron_base.job.executable import Executable
 from pyiron_base.job.jobstatus import JobStatus
 from pyiron_base.job.core import JobCore
-from pyiron_base.generic.util import static_isinstance
+from pyiron_base.job.util import \
+    _copy_restart_files, \
+    _kill_child, \
+    _job_store_before_copy, \
+    _job_reload_after_copy
+from pyiron_base.generic.util import static_isinstance, deprecate
 from pyiron_base.server.generic import Server
 from pyiron_base.database.filetable import FileTable
 import subprocess
-import shutil
 import warnings
 
 __author__ = "Joerg Neugebauer, Jan Janssen"
@@ -463,20 +464,54 @@ class GenericJob(JobCore):
         Returns:
             GenericJob: New GenericJob object pointing to the same job
         """
-        if not self.project_hdf5.file_exists:
-            delete_file_after_copy = True
-        else:
-            delete_file_after_copy = False
-        self.to_hdf()
-        self_class = self.__class__
-        copied_self = self_class(
+        # Store all job arguments in the HDF5 file
+        delete_file_after_copy = _job_store_before_copy(
+            job=self
+        )
+
+        # Copy Python object - super().copy() causes recursion error for serial master
+        copied_self = self.__class__(
             job_name=self.job_name, project=self.project_hdf5.open("..")
         )
-        copied_self.from_hdf()
+        copied_self.reset_job_id()
+
+        # Reload object from HDF5 file
+        _job_reload_after_copy(
+            job=copied_self,
+            delete_file_after_copy=delete_file_after_copy
+        )
+        return copied_self
+
+    def _internal_copy_to(self, project=None, new_job_name=None, new_database_entry=True,
+                          copy_files=True, delete_existing_job=False):
+        # Store all job arguments in the HDF5 file
+        delete_file_after_copy = _job_store_before_copy(
+            job=self
+        )
+
+        # Call the copy_to() function defined in the JobCore
+        new_job_core, file_project, hdf5_project, reload_flag = super(GenericJob, self)._internal_copy_to(
+            project=project,
+            new_job_name=new_job_name,
+            new_database_entry=new_database_entry,
+            copy_files=copy_files,
+            delete_existing_job=delete_existing_job
+        )
+        if reload_flag:
+            return new_job_core, file_project, hdf5_project, reload_flag
+
+        # Reload object from HDF5 file
+        if not static_isinstance(
+            obj=project.__class__,
+            obj_type="pyiron_base.job.core.JobCore"
+        ):
+            _job_reload_after_copy(
+                job=new_job_core,
+                delete_file_after_copy=delete_file_after_copy
+            )
         if delete_file_after_copy:
             self.project_hdf5.remove_file()
-        copied_self._job_id = None
-        return copied_self
+        return new_job_core, file_project, hdf5_project, reload_flag
 
     def copy_to(self, project=None, new_job_name=None, input_only=False, new_database_entry=True,
                 delete_existing_job=False):
@@ -484,7 +519,8 @@ class GenericJob(JobCore):
         Copy the content of the job including the HDF5 file to a new location.
 
         Args:
-            project (ProjectHDFio): The project to copy the job to. (Default is None, use the same project.)
+            project (JobCore/ProjectHDFio/Project/None): The project to copy the job to.
+                (Default is None, use the same project.)
             new_job_name (str): The new name to assign the duplicate job. Required if the project is `None` or the same
                 project as the copied job. (Default is None, try to keep the same name.)
             input_only (bool): [True/False] Whether to copy only the input. (Default is False.)
@@ -495,68 +531,29 @@ class GenericJob(JobCore):
         Returns:
             GenericJob: GenericJob object pointing to the new location.
         """
+        # Update flags
         if input_only and new_database_entry:
             new_database_entry = False
-        if project is None and new_job_name is None:
-            raise ValueError("copy_to requires either a new project or a new_job_name.")
 
-        in_same_project = project is None or project.path == self.project.path
-        has_new_name = new_job_name is not None
-        if in_same_project and not has_new_name:
-            raise ValueError("When copying to the same project, new_job_name must be provided.")
-
-        if not self.project_hdf5.file_exists:
-            self.to_hdf()
-            delete_file_after_copy = True
-        else:
-            delete_file_after_copy = False
-
-        # Get new hdf location
-        project = project or self.project
-        new_job_name = new_job_name or self.job_name
-        job_table = project.job_table(recursive=False)
-        if len(job_table) > 0 and new_job_name in job_table.job.values:
-            if not delete_existing_job:
-                return project.load(new_job_name)
-            else:
-                project.remove_job(new_job_name)
-        if in_same_project and len(self.project_hdf5.h5_path.split("/")) > 2:
-            new_location = self.project_hdf5.open("../" + new_job_name)
-        else:
-            new_location = self.project_hdf5.__class__(project, new_job_name, h5_path="/" + new_job_name)
-
-        # Copy job
-        if in_same_project:
-            new_generic_job = self.copy()
-            new_generic_job.reset_job_id()
-            new_generic_job._name = new_job_name
-            new_generic_job.project_hdf5.copy_to(
-                destination=new_location,
-                maintain_name=False
-            )
-            new_generic_job.project_hdf5 = new_location
-            self._copy_database_entry(
-                new_job_core=new_generic_job,
-                new_database_entry=new_database_entry
-            )
-        else:
-            new_generic_job = super(GenericJob, self).copy_to(
-                project=project,
-                new_database_entry=new_database_entry
-            )
-            new_generic_job.reset_job_id(job_id=new_generic_job.job_id)
-            new_generic_job.from_hdf()
+        # Call the copy_to() function defined in the JobCore
+        new_job_core, file_project, hdf5_project, reload_flag = self._internal_copy_to(
+            project=project,
+            new_job_name=new_job_name,
+            new_database_entry=new_database_entry,
+            copy_files=False,
+            delete_existing_job=delete_existing_job
+        )
+        if reload_flag:
+            return new_job_core
 
+        # Remove output if it should not be copied
         if input_only:
-            if "output" in new_generic_job.project_hdf5.list_groups():
-                del new_generic_job.project_hdf5[
-                    posixpath.join(new_generic_job.project_hdf5.h5_path, "output")
+            if "output" in new_job_core.project_hdf5.list_groups():
+                del new_job_core.project_hdf5[
+                    posixpath.join(new_job_core.project_hdf5.h5_path, "output")
                 ]
-
-        if delete_file_after_copy:
-            self.project_hdf5.remove_file()
-
-        return new_generic_job
+            new_job_core.status.initialized = True
+        return new_job_core
 
     def copy_file_to_working_directory(self, file):
         """
@@ -570,14 +567,15 @@ class GenericJob(JobCore):
         else:
             self.restart_file_list.append(file)
 
-    def copy_template(self, project, new_job_name=None):
+    def copy_template(self, project=None, new_job_name=None):
         """
         Copy the content of the job including the HDF5 file but without the output data to a new location
 
         Args:
-            project (ProjectHDFio): project to copy the job to
-            new_job_name (str): to duplicate the job within the same porject it is necessary to modify the job name
-                                - optional
+            project (JobCore/ProjectHDFio/Project/None): The project to copy the job to.
+                (Default is None, use the same project.)
+            new_job_name (str): The new name to assign the duplicate job. Required if the project is `None` or the same
+                project as the copied job. (Default is None, try to keep the same name.)
 
         Returns:
             GenericJob: GenericJob object pointing to the new location.
@@ -589,31 +587,12 @@ class GenericJob(JobCore):
             new_database_entry=False,
         )
 
-    def _kill_child(self):
-        """
-        Internal helper function to kill a child process.
-        """
-        if not self.server.run_mode.queue and (
-            self.status.running or self.status.submitted
-        ):
-            for proc in psutil.process_iter():
-                try:
-                    pinfo = proc.as_dict(attrs=["pid", "cwd"])
-                except psutil.NoSuchProcess:
-                    pass
-                else:
-                    if pinfo["cwd"] is not None and pinfo["cwd"].startswith(
-                        self.working_directory
-                    ):
-                        job_process = psutil.Process(pinfo["pid"])
-                        job_process.kill()
-
     def remove_child(self):
         """
         internal function to remove command that removes also child jobs.
         Do never use this command, since it will destroy the integrity of your project.
         """
-        self._kill_child()
+        _kill_child(job=self)
         super(GenericJob, self).remove_child()
 
     def kill(self):
@@ -895,30 +874,6 @@ class GenericJob(JobCore):
             "This function needs to be implemented in the specific class."
         )
 
-    # def run_if_non_modal(self):
-    #     """
-    #     The run if non modal function is called by run to execute the simulation in the background. For this we use
-    #     subprocess.Popen()
-    #     """
-    #     shell = (os.name == 'nt')
-    #     try:
-    #         file_name = posixpath.join(self.project_hdf5.working_directory, "run_job.py")
-    #         self._logger.info("{}, status: {}, script: {}".format(self.job_info_str, self.status, file_name))
-    #         with open(posixpath.join(self.project_hdf5.working_directory, 'out.txt'), mode='w') as f_out:
-    #             with open(posixpath.join(self.project_hdf5.working_directory, 'error.txt'), mode='w') as f_err:
-    #                 self._process = subprocess.Popen(['python', '-m', 'pyiron_base.cli', 'wrapper', '-p',
-    #                                                   self.working_directory, '-j', str(self.job_id)],
-    #                                                  cwd=self.project_hdf5.working_directory, shell=shell, stdout=f_out,
-    #                                                  stderr=f_err, universal_newlines=True)
-    #         self._logger.info("{}, status: {}, job submitted".format(self.job_info_str, self.status))
-    #     except subprocess.CalledProcessError as e:
-    #         self._logger.warn("Job aborted")
-    #         self._logger.warn(e.output)
-    #         self.status.aborted = True
-    #         raise ValueError("run_job.py crashed")
-    #     s.logger.info('submitted run %s', self.job_name)
-    #     self._logger.info('job status: %s', self.status)
-
     def run_if_non_modal(self):
         """
         The run if non modal function is called by run to execute the simulation in the background. For this we use
@@ -1233,7 +1188,7 @@ class GenericJob(JobCore):
         if self._check_if_input_should_be_written():
             self.project_hdf5.create_working_directory()
             self.write_input()
-            self._copy_restart_files()
+            _copy_restart_files(job=self)
         self.status.created = True
         self._calculate_predecessor()
         print(
@@ -1350,26 +1305,6 @@ class GenericJob(JobCore):
         if not (self.status.finished or self.status.suspended):
             self.status.aborted = True
 
-    def _copy_restart_files(self):
-        """
-        Internal helper function to copy the files required for the restart job.
-        """
-        if not (os.path.isdir(self.working_directory)):
-            raise ValueError(
-                "The working directory is not yet available to copy restart files"
-            )
-        for i, actual_name in enumerate(
-            [os.path.basename(f) for f in self._restart_file_list]
-        ):
-            if actual_name in self.restart_file_dict.keys():
-                new_name = self.restart_file_dict[actual_name]
-                shutil.copy(
-                    self.restart_file_list[i],
-                    posixpath.join(self.working_directory, new_name),
-                )
-            else:
-                shutil.copy(self.restart_file_list[i], self.working_directory)
-
     def _run_manually(self, _manually_print=True):
         """
         Internal helper function to run a job manually.
@@ -1661,6 +1596,7 @@ class GenericJob(JobCore):
                 self._before_successor_calc(child)
                 child.run()
 
+    @deprecate("Use job.save()")
     def _create_job_structure(self, debug=False):
         """
         Internal helper function to create the input directories, save the job in the database and write the wrapper.
@@ -1668,7 +1604,6 @@ class GenericJob(JobCore):
         Args:
             debug (bool): Debug Mode
         """
-        warnings.warn("Use job.save() instead of job._create_job_structure().", self.s.DeprecationWarning)
         self.save()
 
     def _check_if_input_should_be_written(self):
@@ -1732,10 +1667,3 @@ def multiprocess_wrapper(job_id, working_dir, debug=False, connection_string=Non
         working_directory=str(working_dir), job_id=int(job_id), debug=debug, connection_string=connection_string
     )
     job_wrap.job.run_static()
-
-
-# def multiprocess_master(job_id, working_dir, is_thread_mode=False, debug=False):
-#     job_wrap = JobWrapper(working_directory=str(working_dir), job_id=int(job_id), debug=debug)
-#     job_wrap.job._run_if_refresh()
-#     if is_thread_mode and job_wrap.job._process:
-#         job_wrap.job._process.communicate()
